*************************************************************************************
*														*
*	PDP-8 simulator for EASy68K V1.07a					29/9/2008		*
*														*
*	This, largely untested, code simulates a PDP-8 CPU with 4K of core and		*
*	teletype input/output. It works well enough to run the "Hello world!"		*
*	program that is assembled in the memory as data statements. It will run		*
*	automatically when the simulator is started.						*
*														*
*	This very nearly worked first go. A mistake in hand assembling the PDP-8	*
*	code and a typo in the 68K code prevented the first and second runs from	*
*	completing. Third one's a charm.								*
*														*
*	This version is faster and more complete and works well enough to run DECs	*
*	Octal debugging Tool if it is assembled into memory.					*
*														*
*														*
*	More 68000 and other projects can be found on my website at ..			*
*														*
*	 http://www.themotionstore.com/leeedavison/index.html					*
*														*
*	.. or ..												*
*														*
*	 http://members.lycos.co.uk/leeedavison/index.html					*
*														*
*	mail : leeedavison@lycos.co.uk								*
*														*
*************************************************************************************

	ORG		$1000

start
	MOVEQ		#0,d1				* turn off echo
	MOVEQ		#12,d0			* keyboard echo
	TRAP		#15

	MOVEQ		#0,d3				* zero compare value
	MOVE.w	#$0FFF,d4			* 12 bit mask value

	LEA		core_store(pc),a4		* set the core memory base address. The
							* simulated memory is accessed using the
							* form (a4,dx.w). If more than 16KW is to
							* be simulated this should be changed to
							* (a4,dx.l) to prevent the pointer wrapping
							* round

	MOVE.w	#hello,d6			* set the program counter. The program
							* counter is 12 bits but because it is used
							* as a word address it is made up of b13 to
							* b1 with b0 always being 0

*************************************************************************************

interrupt

* interrupt handling would go in here and be executed before each opcode is executed

*************************************************************************************

opcode_fetch
	ANDI.w	#$1FFE,d6			* mask the PC to 12 bits
	MOVE.w	(a4,d6.w),d0		* get the next opcode

	MOVE.w	d0,-(sp)			* push the opcode
	MOVEQ		#$0E,d1			* set the opcode bits mask
	AND.b		(sp)+,d1			* shift byte and mask the opcode
	CMPI.w	#$000C,d1			* compare with IOT opcode
	BGE.s		no_address			* if IOT or OPR skip the address calculation

* Instruction words are organized as follows:
*	 _ _ _ _ _ _ _ _ _ _ _ _
*	|_|_|_|_|_|_|_|_|_|_|_|_|
*	|     | | |             |
*	|  op |i|z|    addr     |
*
*	op   - the opcode.
*	i    - the indirect bit (0 = direct, 1 = indirect).
*	z    - the page bit (0 = page zero, 1 = current page).
*	addr - the word in page.

* calculate the address. the opcode is in d0.w, the program counter is in d6. leave
* the effective address in d2.w

z_bit		EQU   7				* z bit
i_bit		EQU   8				* i bit

	MOVEQ		#0,d2				* clear the address longword
	BTST		#z_bit,d0			* test the zero page bit
	BEQ.s		zero_page			* if zero page just go use the address

							* else get the page from the program counter
	MOVE.w	#$1F00,d2			* set the address page mask
	AND.w		d6,d2				* mask the page bits to use
zero_page
	MOVE.b	d0,d2				* copy the address from the opcode
	ADD.b		d2,d2				* * 2, the address is a word address

	BTST		#i_bit,d0			* test the indirect bit
	BEQ.s		no_address			* if not set go parse the opcode

* else it is an indirect address so get it

	MOVE.w	(a4,d2.w),d0		* get the indirect address
	CMPI.w	#$001E,d2			* compare address with last autoincrement
							* address
	BGT.s		not_autoincrement		* if greater skip the increment and save

	CMPI.w	#$0010,d2			* compare address with first autoincrement
							* address
	BLT.s		not_autoincrement		* if less skip the increment and save

* else it is an autoincrement address so first increment the address and save the new
* value back

	ADDQ.w	#1,d0				* increment the address
	AND.w		d4,d0				* mask the address to 12 bits
	MOVE.w	d0,(a4,d2.w)		* save the new address value
not_autoincrement
	MOVE.w	d0,d2				* copy the address
	ADD.w		d2,d2				* * 2, the address is a word address
no_address
	ADDQ.w	#2,d6				* increment the PC
	JMP		op_table(pc,d1.w)		* go interpret the opcode

* opcode branch table

op_table
	BRA.s		opcode_0			* opcodes 000 xxx xxx xxx - AND
	BRA.s		opcode_1			* opcodes 001 xxx xxx xxx - TAD
	BRA.s		opcode_2			* opcodes 010 xxx xxx xxx - ISZ
	BRA.s		opcode_3			* opcodes 011 xxx xxx xxx - DCA
	BRA.s		opcode_4			* opcodes 100 xxx xxx xxx - JMS
	BRA.s		opcode_5			* opcodes 101 xxx xxx xxx - JMP
	BRA.s		opcode_6			* opcodes 110 xxx xxx xxx - IOT
	BRA.w		opcode_7			* opcodes 111 xxx xxx xxx - OPR


*************************************************************************************
*
*	000 - AND - and operand with AC
*
* The contents of the memory location are logically anded with the accumulator. There
* is no effect on the link bit.

opcode_0
	MOVE.w	(a4,d2.w),d1		* get the operand value
	AND.w		d4,d1				* mask to 12 bits
	BSET		#12,d1			* set the link bit
	AND.w		d1,d5				* AND the operand with the accumulator
	BRA		interrupt			* opcode done, go test the interrupt flags


*************************************************************************************
*
*	001 - TAD - add operand to <L,AC> (a 13 bit value)
*
* The contents of the memory location are added to the accumulator. A carry out of the
* high bit of the accumulator will complement the link bit.

opcode_1
	MOVE.w	(a4,d2.w),d1		* get the operand value
	AND.w		d4,d1				* mask to 12 bits
	ADD.w		d1,d5				* ADD the operand to the accumulator
	BRA		interrupt			* opcode done, go test the interrupt flags


*************************************************************************************
*
*	010 - ISZ - increment operand and skip if result is zero
*
* The contents of the memory location are incremented. If the result is zero, the
* program counter is incremented; as a result the next instruction will be skipped if
* the result is zero. AC and LINK are not modified. 
*
* ISZ is frequently used to increment memory addresses; unless the address could wrap
* around from 7777 to 0000, ISZ is used as if it will never skip. If a wraparound is
* possible, it must be followed by a no-op

opcode_2
	MOVE.w	(a4,d2.w),d1		* get the value from memory
	ADDQ.w	#1,d1				* + 1
	AND.w		d4,d1				* mask the 12 bit result
	MOVE.w	d1,(a4,d2.w)		* store the value back in memory
	BNE		interrupt			* if not zero result opcode done

	ADDQ.w	#2,d6				* else increment the PC, skip the next opcode
	BRA		interrupt			* opcode done, go test the interrupt flags


*************************************************************************************
*
*	011 - DCA - deposit AC in memory and clear AC
*
* The contents of the accumulator are stored in memory the accumulator is then cleared.
* There is no effect on the link bit

opcode_3
	MOVE.w	d5,d1				* copy the acumulator value
	AND.w		d4,d1				* mask to 12 bits
	MOVE.w	d1,(a4,d2.w)		* store the value in memory
	AND.w		#$1000,d5			* clear the accumulator, link bit unchanged
	BRA		interrupt			* opcode done, go test the interrupt flags


*************************************************************************************
*
*	100 - JMS - jump to subroutine
*
* The program counter is stored in memory location P as a return address, and then
* control is transfered to the location following P. AC and LINK remain unchanged.

opcode_4
	MOVE.w	d6,d1				* copy the program counter
	LSR.w		#1,d1				* shift address bits to b11 - b0
	AND.w		d4,d1				* mask the 12 bits
	MOVE.w	d1,(a4,d2.w)		* store the value in memory

	ADDQ.w	#2,d2				* increment the program address
							* now do a JMP to the address + 1


*************************************************************************************
*
*	101 - JMP - jump
*
* Control is transferred to memory location P.

opcode_5
	MOVE.w	d2,d6				* copy the address to the program counter
	BRA		interrupt			* opcode done, go test the interrupt flags


*************************************************************************************
*
*	110 - IOT - input/output transfer

* The IOT instruction has the following form:
*	 _ _ _ _ _ _ _ _ _ _ _ _
*	|1|1|0|_|_|_|_|_|_|_|_|_|
*	|     |           |     |
*	|     |   device  | op  |

opcode_6
	MOVEQ		#7,d2				* set the mask for the device operation
	AND.w		d0,d2				* mask the device operation
	ADD.w		d2,d2				* * 2 bytes per table entry

	MOVE.w	#$01F8,d1			* set the mask for the device address
	AND.w		d0,d1				* mask the device address

	CMPI.w	#$0020,d1			* compare with TTY write address
	BEQ.s		tty_output			* go handle TTY write operation

	CMPI.w	#$0018,d1			* compare with TTY read address
	BNE		interrupt			* if not read TTY opcode done

* perform a TTY input operation

	JMP		tty_input_table(pc,d2.w)
							* go do the operation

tty_input_table
	BRA.s		tty_kcf			* keyboard clear flags
	BRA.s		tty_ksf			* keyboard skip if flag
	BRA.s		tty_kcc			* keyboard clear and read character
	BRA.s		tty_nop			* keyboard no operation
	BRA.s		tty_krs			* keyboard read static
	BRA.s		tty_kie			* keyboard interrupt enable
	BRA.s		tty_krb			* keyboard read and begin next
tty_nop
	BRA		interrupt			* opcode done, go test the interrupt flags


*------------------------------------------------------------------------------------
*
* keyboard clear flags. The keyboard flag, signalling input data ready, is cleared.

tty_kcf
	MOVEQ		#7,d0				* read keyboard flag
	TRAP		#15

	TST.b		d1				* test the flag
	BEQ.s		no_flag			* skip read if no flag

	MOVEQ		#5,d0				* read the keyboard buffer, clear the flag
	TRAP		#15

no_flag
	BRA		interrupt			* opcode done, go test the interrupt flags


*------------------------------------------------------------------------------------
*
* keyboard skip if flag. If the keyboard flag is set, indicating that input data is
* ready, the next instruction in sequence is skipped

tty_ksf
	MOVEQ		#7,d0				* read keyboard flag
	TRAP		#15

	TST.b		d1				* test the flag
	BEQ		interrupt			* if no flag opcode done

	ADDQ.w	#2,d6				* else increment the program counter
	BRA		interrupt			* opcode done, go test the interrupt flags


*------------------------------------------------------------------------------------
*
* keyboard clear and read character. The keyboard flag is reset, the accumulator is
* cleared, and the process of reading the next character of input is initiated

tty_kcc
	AND.w		#$1000,d5			* clear the accumulator
	BRA		interrupt			* opcode done, go test the interrupt flags


*------------------------------------------------------------------------------------
*
* keyboard interrupt enable. The accumulator is loaded into the device control register

tty_kie
	BRA		interrupt			* opcode done, go test the interrupt flags


*------------------------------------------------------------------------------------
*
* keyboard read and begin next. The 8 bit character in the keyboard buffer is
* transferred to the accumulator, and the keyboard flag is cleared, allowing the
* reading of the next character to begin. In effect, this operation combines the KCC
* and KRS operations

tty_krb
	AND.w		#$1000,d5			* clear the accumulator

*------------------------------------------------------------------------------------
*
* keyboard read static. The 8-bit character in the keyboard buffer is ored with the
* accumulator
*
* Most early PDP-8 systems used teletypes as console terminals; as sold by DEC, these
* were configured for mark parity, so most older software assumes 7 bit ASCII, upper
* case only, with the 8th bit set to 1

tty_krs
	MOVEQ		#5,d0				* read the keyboard buffer, clear the flag
	TRAP		#15

	OR.b		d1,d5				* OR the key with the accumulator
	BSET		#7,d5				* set the eighth nit
	BRA		interrupt			* opcode done, go test the interrupt flags


*------------------------------------------------------------------------------------
*------------------------------------------------------------------------------------
*
* perform a TTY output operation

tty_output
	JMP		tty_output_table(pc,d2.w)
							* go do the operation

tty_output_table
	BRA.s		tty_tfl			* teleprinter flag set
	BRA.s		tty_tsf			* teleprinter skip if flag
	BRA.s		tty_tcf			* teleprinter clear flag
	BRA.s		tty_nop			* teleprinter no operation
	BRA.s		tty_tpc			* teleprinter print character
	BRA.s		tty_tsk			* teleprinter skip
	BRA.s		tty_tls			* teleprinter load and start
	BRA		interrupt			* opcode done, go test the interrupt flags


*------------------------------------------------------------------------------------
*
* teleprinter flag set. The printer flag, signalling output complete, is set.

tty_tfl
	BRA		interrupt			* opcode done, go test the interrupt flags


*------------------------------------------------------------------------------------
*
* teleprinter skip if flag. If the printer flag is set, indicating output is complete,
* the next instruction in sequence is skipped. As output in EASy68K is always complete
* the next instruction is always skipped

tty_tsf
	ADDQ.w	#2,d6				* increment the program counter
	BRA		interrupt			* opcode done, go test the interrupt flags


*------------------------------------------------------------------------------------
*
* teleprinter clear flag. The printer flag is reset

tty_tcf
	BRA		interrupt			* opcode done, go test the interrupt flags


*------------------------------------------------------------------------------------
*
* teleprinter load and start. The least significant 8 bits of the accumulator are
* copied to the print buffer, initiating output, and the printer flag is reset. In
* effect, this operation combines the TCF and TPC operations. As TCF is essentially
* a NOP on EASy68K this makes TLS and TPC identical so the same code is used for both

tty_tls

*------------------------------------------------------------------------------------
*
* teleprinter print character. The least significant 8-bits of the accumulator is
* copied to the print buffer, initiating output.
*
* Most early PDP-8 systems used teletypes as console terminals; as sold by DEC, these
* were configured for mark parity, so most older software assumes 7 bit ASCII, upper
* case only, with the 8th bit set to 1

tty_tpc
	MOVEQ		#$7F,d1			* set the character mask
	AND.b		d5,d1				* mask the character to send
	MOVEQ		#6,d0				* output character
	TRAP		#15

	BRA		interrupt			* opcode done, go test the interrupt flags


*------------------------------------------------------------------------------------
*
* teleprinter skip. If either the print flag or the keyboard flag are set, the next
* instruction in sequence is skipped.

tty_tsk
	ADDQ.w	#2,d6				* increment the program counter
	BRA		interrupt			* opcode done, go test the interrupt flags


*************************************************************************************
*
*	111 - OPR - microcoded operations

CLA		EQU 7					* clear AC
CLL		EQU 6					* clear the L bit
CMA		EQU 5					* ones complement AC
CML		EQU 4					* complement L bit
IAC		EQU 0					* increment <L,AC>

group_2	EQU 8					* group two microcoded opcode bit

opcode_7
	BTST		#group_2,d0			* test the group 2 bit
	BNE		opcode_7_2			* if set go do group 2 opcodes

*          _ _ _ _ _ _ _ _ _ _ _ _
* Group 1 |1|1|1|0|_|_|_|_|_|_|_|_|
*                  1                - CLA - clear AC
*                    1              - CLL - clear the L bit
*                      1            - CMA - ones complement AC
*                        1          - CML - complement L bit
*                                1  - IAC - increment <L,AC>
*                          0 0 1    - BSW - byte swap
*                          1 0 0    - RAR - rotate <L,AC> right
*                          0 1 0    - RAL - rotate <L,AC> left
*                          1 1 0    - RAN - rotate <L,AC> nowhere
*                          1 0 1    - RTR - rotate <L,AC> right twice
*                          0 1 1    - RTL - rotate <L,AC> left twice
*                          1 1 1    - RAN - rotate <L,AC> nowhere twice
*
* In general, the above operations can be combined by oring the bit patterns for the
* desired operations into a single instruction. If none of the bits are set, the
* result is the NOP instruction. When these operations are combined, they operate top
* to bottom in the order shown above. The exception to this is that IAC cannot be
* combined with the rotate operations on some models, and attempts to combine rotate
* operations have different effects from one model to another (for example, on the
* PDP-8/E, the rotate code 001 means swap 6 bit bytes in the accumulator, while
* previous models took this to mean something like "shift neither left nor right 2
* bits").

	BTST		#CLA,d0			* test the CLA bit
	BEQ.s		not_CLA			* if not set skip CLA

	ANDI.w	#$1000,d5			* clear the accumulator
not_CLA
	BTST		#CLL,d0			* test the CLL bit
	BEQ.s		not_CLL			* if not set skip CLL

	AND.w		d4,d5				* clear the link bit
not_CLL
	BTST		#CMA,d0			* test the CMA bit
	BEQ.s		not_CMA			* if not set skip CMA

	EOR.w		d4,d5				* complement the accumulator
not_CMA
	BTST		#CML,d0			* test the CML bit
	BEQ.s		not_CML			* if not set skip CML

	EORI.w	#$1000,d5			* complement the link bit
not_CML
	BTST		#IAC,d0			* test the IAC bit
	BEQ.s		not_IAC			* if not set skip IAC

	ADDQ.w	#1,d5				* increment the accumulator
	ANDI.w	#$1FFF,d5			* mask the acumulator and link bit
not_IAC
	MOVEQ		#$0E,d2			* set mask for rotates and swap
	AND.w		d0,d2				* mask rotate bits
	BEQ		interrupt			* 000 opcode done, go test the interrupt flags	

	ADD.b		d2,d2				* * 2, 4 bytes per rotate table entry
	JMP		rotate_table(pc,d2.w)	* go interpret the opcode

rotate_table
	BRA.w		interrupt			* 000 opcode done, go test the interrupt flags	
	BRA.w		byte_swap			* 001 byte swap
	BRA.w		rotate_left			* 010 rotate left
	BRA.w		rotate_left_twice		* 011 rotate left twice
	BRA.w		rotate_right		* 100 rotate right
	BRA.w		rotate_right_twice	* 101 rotate right twice
	BRA.w		interrupt			* 110 opcode done, go test the interrupt flags	
	BRA.w		interrupt			* 111 opcode done, go test the interrupt flags	


*------------------------------------------------------------------------------------
*
* 001 byte swap
*
* swap the high 6 bits of the accumulator with the low 6 bits of the accumulator

byte_swap
	MOVE.w	d5,d2				* copy the accumulator
	MOVE.w	d5,d1				* copy the accumulator again
	ANDI.w	#$003F,d1			* mask the low byte
	ANDI.w	#$0FC0,d2			* mask the high byte
	ANDI.w	#$1000,d5			* clear the accumulator
	LSR.w		#6,d2				* shift the high byte to the low byte
	LSL.w		#6,d1				* shift the low byte to the high byte
	OR.w		d1,d5				* OR the shifetd low byte into the accumulator
	OR.w		d2,d5				* OR the shifetd high byte into the accumulator
	BRA		interrupt			* opcode done, go test the interrupt flags	


*------------------------------------------------------------------------------------
*
* 010 rotate left
*
* rotate the 13 bit LINK/AC register one place left, so that the most significant bit
* of AC becomes the new link, and the old link becomes the least significant bit of AC

rotate_left
	LSL.w		#1,d5				* shift right one bit
	MOVE.w	d5,d2				* copy the result
	ROL.w		#3,d2				* shift b13 to b0
	ANDI.w	#1,d2				* mask b0
	OR.w		d2,d5				* OR into the accumulator word
	ANDI.w	#$1FFF,d5			* mask the acumulator and link bit
	BRA		interrupt			* opcode done, go test the interrupt flags	


*------------------------------------------------------------------------------------
*
* 011 rotate left twice
*
* equivalent to two consecutive rotate left instructions

rotate_left_twice
	LSL.w		#2,d5				* shift right two bits
	MOVE.w	d5,d2				* copy the result
	ROL.w		#3,d2				* shift b14, b13 to b1, b0
	ANDI.w	#3,d2				* mask b1, b0
	OR.w		d2,d5				* OR into the accumulator word
	ANDI.w	#$1FFF,d5			* mask the acumulator and link bit
	BRA		interrupt			* opcode done, go test the interrupt flags	


*------------------------------------------------------------------------------------
*
* 100 rotate right
*
* rotate the 13 bit LINK/AC register one place right, so that the least significant bit
* of AC becomes the new link, and the old link becomes the most significant bit of AC

rotate_right
	MOVEQ		#1,d2				* set mask for b1, b0
	AND.w		d5,d2				* copy the accumulator b1, b0
	ROR.w		#3,d2				* shift b0 to b13
	OR.w		d2,d5				* OR into the accumulator word
	LSR.w		#1,d5				* shift right one bit
	ANDI.w	#$1FFF,d5			* mask the acumulator and link bit
	BRA		interrupt			* opcode done, go test the interrupt flags	


*------------------------------------------------------------------------------------
*
* 101 rotate right twice
*
* equivalent to two consecutive rotate right instructions

rotate_right_twice
	MOVEQ		#3,d2				* set mask for b1, b0
	AND.w		d5,d2				* copy the accumulator b1, b0
	ROR.w		#3,d2				* shift b1, b0 to b14, b13
	OR.w		d2,d5				* OR into the accumulator word
	LSR.w		#2,d5				* shift right two bits
	ANDI.w	#$1FFF,d5			* mask the acumulator and link bit
	BRA		interrupt			* opcode done, go test the interrupt flags	


*************************************************************************************
*
*	111 1 - OPR - group 2 microcoded operations

*          _ _ _ _ _ _ _ _ _ _ _ _
* Group 2 |1|1|1|1|_|_|_|_|_|_|_|0|
*                    1     0        - SMA - skip on AC < 0  \
*                      1   0        - SZA - skip on AC = 0   > OR group
*                        1 0        - SNL - skip on L /= 0  /
*                    0 0 0 1        - SKP - skip unconditionally
*                    1     1        - SPA - skip on AC >= 0 \
*                      1   1        - SNA - skip on AC /= 0  > AND group
*                        1 1        - SZL - skip on L = 0   /
*                  1                - CLA - clear AC
*                            1      - OSR - or switches with AC
*                              1    - HLT - halt
*
* The above operations may be combined by oring them together, except that there are
* two distinct incompatible groups of skip instructions. When combined, SMA, SZA and
* SNL, skip if one or the other of the indicated conditions are true (logical or),
* while SPA, SNA and SZL skip if all of the indicated conditions are true (logical
* and). When combined, these operate top to bottom in the order shown; thus, the
* accumulator may be tested and then cleared. Setting the halt bit in a skip
* instruction is a crude but useful way to set a breakpoint for front-panel debugging.
* If none of the bits are set, the result is an alternative form of no-op.

opcode_7_2
	BTST		#0,d0				* test bit 0
	BNE		interrupt			* opcode done, go test the interrupt flags

	MOVE.w	d4,d1				* set the accumulator mask

	MOVEQ		#$78,d2			* set mask for skip bits
	AND.w		d0,d2				* mask skip bits
	ROR.w		#2,d2				* / 4, two bytes per table entry
	JMP		skip_table(pc,d2.w)	* go do skip opcode

skip_table
	BRA.s		group_2_tail		* 0000 no skip
	BRA.s		set_skip			* 0001 unconditional skip
	BRA.s		snl_skip			* 0010 skip on L <> 0
	BRA.s		szl_skip			* 0011 skip on L = 0
	BRA.s		sza_skip			* 0100 skip on accumulator = 0
	BRA.s		sna_skip			* 0101 skip on accumulator <> 0
	BRA.s		sza_snl_skip		* 0110 skip on accumulator = 0 OR L <> 0
	BRA.s		sna_szl_skip		* 0111 skip on accumulator /= 0 AND L = 0
	BRA.s		sma_skip			* 1000 skip on accumulator < 0
	BRA.s		spa_skip			* 1001 skip on accumulator >= 0
	BRA.s		sma_snl_skip		* 1010 skip on accumulator < 0 OR L <> 0
	BRA.s		spa_szl_skip		* 1011 skip on accumulator >= 0 AND L = 0
	BRA.s		sma_sza_skip		* 1100 skip on accumulator <= 0
	BRA.s		spa_sna_skip		* 1101 skip on accumulator > 0
	BRA.s		sma_sza_snl_skip		* 1110 skip on accumulator <= 0 OR L <> 0
*##	BRA.s		spa_sna_szl_skip		* 1111 skip on accumulator > 0 AND L = 0

link		EQU 12				* link bit
sign		EQU 11				* sign bit

*------------------------------------------------------------------------------------
*
* 1111 skip on accumulator > 0 AND L = 0

spa_sna_szl_skip
	AND.w		d5,d1				* mask the accumulator
	CMP.w		d3,d1				* compare with zero
	BGT.s		szl_skip			* if > 0 go test L

	BRA.s		group_2_tail		* else just do tail


*------------------------------------------------------------------------------------
*
* 1011 skip on accumulator >= 0 AND L = 0

spa_szl_skip
	BTST		#sign,d5			* test the accumulator sign bit
	BNE.s		group_2_tail		* if negative just do tail

*------------------------------------------------------------------------------------
*
* 0011 skip on L = 0

szl_skip
	BTST		#link,d5			* test the link bit
	BEQ.s		set_skip			* if not set go do skip

	BRA.s		group_2_tail		* else just do tail


*------------------------------------------------------------------------------------
*
* 0100 skip on accumulator = 0

sza_skip
	AND.w		d5,d1				* mask the accumulator
	BEQ.s		set_skip			* if clear go do skip

	BRA.s		group_2_tail		* else just do tail


*------------------------------------------------------------------------------------
*
* 0101 skip on accumulator <> 0

sna_skip
	AND.w		d5,d1				* mask the accumulator
	BNE.s		set_skip			* if not clear go do skip

	BRA.s		group_2_tail		* else just do tail


*------------------------------------------------------------------------------------
*
* 0111 skip on accumulator /= 0 AND L = 0

sna_szl_skip
	AND.w		d5,d1				* mask the accumulator
	BNE.s		szl_skip			* if not clear go test L

	BRA.s		group_2_tail		* else just do tail


*------------------------------------------------------------------------------------
*
* 1010 skip on accumulator < 0 OR L <> 0

sma_snl_skip
	BTST		#link,d5			* test the link bit
	BNE.s		set_skip			* if set go do skip

*------------------------------------------------------------------------------------
*
* 1000 skip on accumulator < 0

sma_skip
	BTST		#sign,d5			* test the accumulator sign bit
	BNE.s		set_skip			* if negative go do skip

	BRA.s		group_2_tail		* else just do tail


*------------------------------------------------------------------------------------
*
* 1001 skip on accumulator >= 0

spa_skip
	BTST		#sign,d5			* test the accumulator sign bit
	BEQ.s		set_skip			* if positive go do skip

	BRA.s		group_2_tail		* else just do tail


*------------------------------------------------------------------------------------
*
* 1100 skip on accumulator <= 0

sma_sza_skip
	AND.w		d5,d1				* mask the accumulator
	CMP.w		d3,d1				* compare with zero
	BLE.s		set_skip			* if <= 0 go do skip

	BRA.s		group_2_tail		* else just do tail


*------------------------------------------------------------------------------------
*
* 1101 skip on accumulator > 0

spa_sna_skip
	AND.w		d5,d1				* mask the accumulator
	CMP.w		d3,d1				* compare with zero
	BGT.s		set_skip			* if > 0 go do skip

	BRA.s		group_2_tail		* else just do tail


*------------------------------------------------------------------------------------
*
* 1110 skip on accumulator <= 0 OR L <> 0

sma_sza_snl_skip
	AND.w		d5,d1				* mask the accumulator
	CMP.w		d3,d1				* compare with zero
	BGT.s		snl_skip			* if > 0 go test L

	BRA.s		group_2_tail		* else just do tail


*------------------------------------------------------------------------------------
*
* 0110 skip on accumulator = 0 OR L <> 0

sza_snl_skip
	AND.w		d5,d1				* mask the accumulator
	BEQ.s		set_skip			* if clear go do skip

*------------------------------------------------------------------------------------
*
* 0010 skip on L <> 0

snl_skip
	BTST		#link,d5			* test the link bit
	BEQ.s		group_2_tail		* if not set just do tail

*------------------------------------------------------------------------------------
*
* 0001 unconditional skip
* set the skip offset

set_skip
	ADDQ.w	#2,d6				* and increment the PC

*------------------------------------------------------------------------------------
*
* 0000 no skip
* continue group 2 processing

CLA		EQU 7					* clear AC bit
OSR		EQU 2					* OR switches with AC
HLT		EQU 1					* halt

group_2_tail
	BTST		#CLA,d0			* test the clear bit
	BEQ.s		no_clear_AC			* if not set skip the clear

	AND.w		#$1000,d5			* else clear the accumulator
no_clear_AC
*##	BTST		#OSR,d0			* test the OR switches bit
*##	BEQ.s		no_OR_switches		* if not set skip the OR
*##
*##	OR.w		#$0000,d5			* else OR switches with AC
*##no_OR_switches
	BTST		#HLT,d0			* test the halt bit
	BEQ.s		no_halt			* if not set skip the halt

	MOVEQ		#9,d0				* halt simulator
	TRAP		#15

no_halt
	BRA		interrupt			* opcode done, go test the interrupt flags


*************************************************************************************
*
* simulated core store. this core has a "Hello world!" program pre loaded into it and
* the start address is automatically loaded into the program counter at startup.
*
* The memory is only 12 bits wide, the upper 4 bits being unused.

	ORG	$4000
core_store
	ds.w	$08					* skip the first eight words
ptr						EQU	*-core_store
	dc.w	string/2				* word $08, the first autoincrement register

	ds.w	$77					* skip the rest of this page

hello						EQU	*-core_store
	dc.w	$0EC0					* CLA, CLL
	dc.w	$0300+ptr/2				* TAD I Z ptr
	dc.w	$0F28					* SNA
	dc.w	$0F02					* HLT
	dc.w	$0C26					* TLS
llp						EQU	*-core_store
	dc.w	$0C21					* TSF
	dc.w	$0A00+llp/2				* JMP .-1
	dc.w	$0A00+hello/2			* JMP hello

* Most early PDP-8 systems used teletypes as console terminals; as sold by DEC, these
* were configured for mark parity, so most older software assumes 7 bit ASCII, upper
* case only, with the 8th bit set to 1

string					EQU	*-2-core_store
	dc.b	0,'H'+$80				* 
	dc.b	0,'e'+$80				* 
	dc.b	0,'l'+$80				* 
	dc.b	0,'l'+$80				* 
	dc.b	0,'o'+$80				* 
	dc.b	0,' '+$80				* 
	dc.b	0,'w'+$80				* 
	dc.b	0,'o'+$80				* 
	dc.b	0,'r'+$80				* 
	dc.b	0,'l'+$80				* 
	dc.b	0,'d'+$80				* 
	dc.b	0,'!'+$80				* 
	dc.w	0					* end marker


*************************************************************************************

	END		start

*************************************************************************************
*
*	Notes cribbed from http://www.faqs.org/faqs/dec-faq/pdp8/
*
*************************************************************************************
*
* The PDP-8 word size is 12 bits, and the basic memory is 4K words. The minimal CPU
* contained the following registers:
*
*	PC - the program counter, 12 bits.
*	AC - the accumulator, 12 bits.
*	L  - the link, 1 bit, commonly prefixed to AC as <L,AC>.
*
* It is worth noting that many operations such as procedure linkage and indexing,
* which are usually thought of as involving registers, are done with memory on the
* PDP-8 family.
*
* Instruction words are organized as follows:
*	 _ _ _ _ _ _ _ _ _ _ _ _
*	|_|_|_|_|_|_|_|_|_|_|_|_|
*	|     | | |             |
*	|  op |i|z|    addr     |
*
*	op   - the opcode.
*	i    - the indirect bit (0 = direct, 1 = indirect).
*	z    - the page bit (0 = page zero, 1 = current page).
*	addr - the word in page.
*
* The top 5 bits of the 12 bit program counter give the current page, and memory
* addressing is also complicated by the fact that absolute memory locations 8 through
* 15 are incremented prior to use when used as indirect addresses. These locations
* are called auto-index registers (despite the fact that they are in memory); they
* allow the formulation of very tightly coded array operations.
*
* The basic instructions are:
*
*	000 - AND - and operand with AC.
*	001 - TAD - add operand to <L,AC> (a 13 bit value).
*	010 - ISZ - increment operand and skip if result is zero.
*	011 - DCA - deposit AC in memory and clear AC.
*	100 - JMS - jump to subroutine.
*	101 - JMP - jump.
*	110 - IOT - input/output transfer.
*	111 - OPR - microcoded operations.
*
* The ISZ and other skip instructions conditionally skip the next instruction in
* sequence. The ISZ is commonly used to increment a loop counter and skip if done,
* and it is also used as an general increment instruction, either followed by a no-op
* or in contexts where it is known that the result will never be zero.
*
* The JMS instruction stores the return address in relative word zero of the
* subroutine, with execution starting with relative word one. Subroutine return is
* done with an indirect JMP through the return address. Subroutines commonly increment
* their return addresses to index through inline parameter lists or to perform
* conditional skips over instructions following the call.
*
* The IOT instruction has the following form:
*	 _ _ _ _ _ _ _ _ _ _ _ _
*	|1|1|0|_|_|_|_|_|_|_|_|_|
*	|     |           |     |
*	|     |   device  | op  |
*
* The IOT instruction specifies one of up to 8 operations on one of 64 devices.
* Typically (but not universally), each bit of the op field evokes an operation, and
* these can be microcoded in right to left order. Prior to the PDP-8/E, there were
* severe restrictions on the interpretation of the op field that resulted from the
* fact that the operation was delivered as a sequence of IOP pulses, each on a separate
* line of the I/O bus. Each line was typically used to evoke a different device
* function, so essentially, the operation 000 was always a no-op because it evoked no
* functions, and the code 111 evoked all three functions in series.
*
* As an example of the use of IOT instructions, consider the console terminal
* interface. On early PDP-8 systems, this was always assumed to be an ASR 33 teletype,
* complete with low-speed paper tape reader and punch. It was addressed as devices 03
* (the keyboard/reader) and 04 (the teleprinter/punch):
*	 _ _ _ _ _ _ _ _ _ _ _ _
*	|1|1|0|_|_|_|_|_|_|_|_|_|
*	      |0 0 0 0 1 1|0 0 1  - KSF - keyboard skip if flag
*	      |0 0 0 0 1 1|0 1 0  - KCC - keyboard clear flag
*	      |0 0 0 0 1 1|1 0 0  - KRS - keyboard read static
*
* The keyboard flag is set by the arrival of a character. The KCC instruction clears
* both the flag and the accumulator. KRS ors the 8 bit input data with the low order
* 8 bits of AC. The commonly used KRB instruction is the or of KCC and KRS. To await
* one byte of input, use KSF to poll the flag, then read the byte with KRB.
*	 _ _ _ _ _ _ _ _ _ _ _ _
*	|1|1|0|_|_|_|_|_|_|_|_|_|
*	      |0 0 0 1 0 0|0 0 1  - TSF - teleprinter skip if flag
*	      |0 0 0 1 0 0|0 1 0  - TCF - teleprinter clear flag
*	      |0 0 0 1 0 0|1 0 0  - TPC - teleprinter print static
*
* The teleprinter flag is set by the completion of the TPC operation (as a result, on
* startup, many applications output a null in order to get things going). TCF clears
* the flag, and TPC outputs the low order 8 bits of the accumulator. The commonly
* used TLS instruction is the or of TCF and TPC. To output a character, first use TSF
* to poll the flag, then write the character with TLS.
*
* IOT instructions may be used to initiate data break transfers from block devices
* such as disk or tape. The term "data break" was, for years, DEC's preferred term
* for cycle-stealing direct-memory-access data transfers.
*
* Some CPU functions are accessed only by IOT instructions. For example, interrupt
* enable and disable are IOT instructions:
*	 _ _ _ _ _ _ _ _ _ _ _ _
*	|1|1|0|_|_|_|_|_|_|_|_|_|
*	      |0 0 0 0 0 0|0 0 1  - ION - interrupts turn on
*	      |0 0 0 0 0 0|0 1 0  - IOF - interrupts turn off
*
* An interrupt is requested when any device raised its flag. The console master clear
* switch resets all flags and disables interrupts. In effect, an interrupt is a JMS
* instruction to location zero, with the side effect of disabling interrupts. The
* interrupt service routine is expected to test the device flags and perform the
* operations needed to reset them, and then return using ION immediately before the
* indirect return JMP. The effect of ION is delayed so that interrupts are not enabled
* until after the JMP.
*
* The instructions controlling the optional memory management unit are also IOT
* instructions. This unit allows the program to address up to 32K of main memory by
* adding a 3 bit extension to the memory address. Two extensions are available, one
* for instruction fetch and direct addressing, the other for indirect addressing.
*
* A wide variety of operations are available through the OPR microcoded instructions:
*          _ _ _ _ _ _ _ _ _ _ _ _
* Group 1 |1|1|1|0|_|_|_|_|_|_|_|_|
*                  1                - CLA - clear AC
*                    1              - CLL - clear the L bit
*                      1            - CMA - ones complement AC
*                        1          - CML - complement L bit
*                                1  - IAC - increment <L,AC>
*                          1 0 0    - RAR - rotate <L,AC> right
*                          0 1 0    - RAL - rotate <L,AC> left
*                          1 0 1    - RTR - rotate <L,AC> right twice
*                          0 1 1    - RTL - rotate <L,AC> left twice
*
* In general, the above operations can be combined by oring the bit patterns for the
* desired operations into a single instruction. If none of the bits are set, the
* result is the NOP instruction. When these operations are combined, they operate top
* to bottom in the order shown above. The exception to this is that IAC cannot be
* combined with the rotate operations on some models, and attempts to combine rotate
* operations have different effects from one model to another (for example, on the
* PDP-8/E, the rotate code 001 means swap 6 bit bytes in the accumulator, while
* previous models took this to mean something like "shift neither left nor right 2
* bits").
*          _ _ _ _ _ _ _ _ _ _ _ _
* Group 2 |1|1|1|1|_|_|_|_|_|_|_|0|
*                    1     0        - SMA - skip on AC < 0  \
*                      1   0        - SZA - skip on AC = 0   > OR group
*                        1 0        - SNL - skip on L /= 0  /
*                    0 0 0 1        - SKP - skip unconditionally
*                    1     1        - SPA - skip on AC >= 0 \
*                      1   1        - SNA - skip on AC /= 0  > AND group
*                        1 1        - SZL - skip on L = 0   /
*                  1                - CLA - clear AC
*                            1      - OSR - or switches with AC
*                              1    - HLT - halt
*
* The above operations may be combined by oring them together, except that there are
* two distinct incompatible groups of skip instructions. When combined, SMA, SZA and
* SNL, skip if one or the other of the indicated conditions are true (logical or),
* while SPA, SNA and SZL skip if all of the indicated conditions are true (logical
* and). When combined, these operate top to bottom in the order shown; thus, the
* accumulator may be tested and then cleared. Setting the halt bit in a skip
* instruction is a crude but useful way to set a breakpoint for front-panel debugging.
* If none of the bits are set, the result is an alternative form of no-op.
*
* A third group of operate microinstructions (with a 1 in the least significant bit)
* deals with the optional extended arithmetic element to allow such things as hardware
* multiply and divide, 24 bit shift operations, and normalize. These operations
* involve an additional data register, MQ or multiplier quotient, and a small step
* count register. On the PDP-8/E and successors, MQ and the instructions for loading
* and storing it were always present, even when the EAE was absent, and the EAE was
* extended to provide a useful variety of 24 bit arithmetic operations.
*
*************************************************************************************
*
* There are many different assemblers for the PDP-8, but most use a compatible basic
* syntax; here is an example:
*
*	START,	CLA CLL		/ Clear everything
*		TAD	X	/ Load X
*		AND I	Y	/ And with the value pointed to by Y
*		DCA	X	/ Store in X
*		HLT		/ Halt
*
*	X,	1		/ A variable
*	Y,	7		/ A pointer
*
* Note that labels are terminated by a comma, and comments are separated from the
* code by a slash. There are no fixed fields or column restrictions. The "CLA CLL"
* instruction on the first line is an example of the microcoding of two of the Group
* 1 operate instructions. CLA alone has the code 7200 (octal), while CLL has the code
* 7100; combining these as "CLA CLL" produces 7300. As a general rule, except when
* memory reference instructions are involved, the assembler simply ors together the
* values of all blank separated fields between the label and comment.
*	
* Indirection is indicated by the special symbol I in the operand field, as in the
* third line of the example. The typical PDP-8 assembler has no explicit notation to
* distinguish between page zero and current page addresses. Instead, the assembler is
* expected to note the page holding the operand and automatically generate the
* appropriate mode. If the operand is neither in the current page nor page zero, some
* assemblers will raise an error, others will automatically generate an indirect
* pointer to the off-page operand; this should be avoided because it only works for
* directly addressed off-page operands, and only when the memory management unit is
* not being used to address a data field other than the current instruction field.
*
* Note, in the final two lines of the example, that there is no "define constant"
* pseudo-operation. Instead, where a constant is to be assembled into memory, the
* constant takes the place of the op-code field.
*
* The PDP-8 has no immediate addressing mode, but most assemblers provide a notation
* to allow the programmer to ignore this lack:
*
*		TAD	(3)	/ add 3, from memory on the current page.
*		TAD	[5]	/ add 5, from memory on page zero.
*		JMP I	(LAB)	/ jump indirect through the address of LAB.
*
* Assemblers that support this automatically fill the end of each page with constants
* defined in this way that have been accumulated during the assembly of that page.
* Note that the variants "(3" and "[5" (with no closing parentheses) are usually
* allowed but the use of this sloppy form is discouraged. Furthermore, the widely
* used PAL8 assembler interprets the unlikely operand "(3)+1" as being the same as
* "(3+1)".
*
* Arithmetic is allowed in operand fields and constant definitions, with expressions
* evaluated in strict left-to-right order, as:
*
*		TAD	X+1	/ add the contents of the location after X.
*		TAD	(X-1)	/ add the address of the location before X.
*
* Other operators allowed include and (&), or (!), multiply (^) and divide (%), as
* well as a unary sign (+ or -). Unfortunately, one of the most widely used
* assemblers, PAL8, has trouble when unary operators are mixed with multiplication
* or division.
*	
* Generally, only the first 6 characters of identifiers are significant and numeric
* constants are evaluated in octal.
*
* Other assembly language features are illustrated below:
*
*	/ Comments may stand on lines by themselves
*				/ Blank lines are allowed
*
*		*200		/ Set the assembly origin to 200 (octal)
*
*	NL0002=	CLA CLL CML RTL	/ Define new opcode NL0002.
*
*		NL0002	/ Use new opcode (load 0002 in AC)
*		JMP	.-1	/ Jump to the previous instruction
*
*	X1=	10		/ Define X1 (an auto-index register address)
*	LETA=	"A		/ Define LETA as 000011000001 (ASCII A)
*
*		TAD I	X1	/ Use autoindex register 1
*
*		IAC; RAL	/ Multiple instructions on one line
*
*		$		/ End of assembly
*
* The assembly file ends with a line containing a $ (dollar sign) not in a comment
* field.
*
* The $, * and =  syntax used by most PDP-8 assemblers replaces functions performed
* by pseudo-operations on many other assemblers. In addition, PAL8, the most widely
* used PDP-8 assembler supports the following pseudo-operations:
*
*		DECIMAL		/ Interpret numeric constants in base 10
*		OCTAL			/ Interpret numeric constants in base 8
*		EJECT			/ Force a page eject in the listing
*		XLIST			/ Toggle listing
*		XLIST	N		/ Turn on listing if N=0, off if N=1
*		PAGE			/ Advance location counter to next page
*		PAGE	N		/ Set location counter start of page N
*		FIELD	N		/ Assemble into extended memory field N
*		TEXT	"STR"		/ Pack STR into consecutive 6 bit bytes
*		ZBLOCK	N	/ Allocate N words, initialized to zero
*		IFDEF	S <C>		/ Assemble C if symbol S is defined
*		IFNDEF	S <C>	/ Assemble C if symbol S is not defined
*		IFZERO	E <C>	/ Assemble C if expression E is zero
*		IFNZRO	E <C>	/ Assemble C if expression E is not zero
*		FIXMRI  OP= VAL	/ Define OP as memory reference instruction
*
* Conditonally assembled code must be enclosed in angle brackets. The enclosed code
* may extend over multiple lines and, because different assemblers treat comments
* within conditionals differently, the closing bracket should not be in a comment and
* any brackets in comments should be balanced.
*
*************************************************************************************
*
* PDP-8/I boot loader, allegedly
*
* addr word
* ---- ----
* 7756 6032  - KCC - keyboard clear flag
* 7757 6031  - KSF - keyboard skip if flag
* 7760 5357  - JMP - jump, direct, zero page, addr x157
*
* 7761 6036  - KRB - keyboard clear flag, read static
* 7762 7106  - CLL - clear the L bit
*		 - RTL - rotate <L,AC> left twice
* 7763 7006  - RTL - rotate <L,AC> left twice
* 7764 7510  - SPA - skip on AC >= 0
* 7765 5357  - JMP - jump, direct, zero page, addr x157
*
* 7766 7006  - RTL - rotate <L,AC> left twice
* 7767 6031  - KSF - keyboard skip if flag
* 7770 5367  - JMP - jump, direct, zero page, addr x167
*
* 7771 6034  - KRS - keyboard read static
* 7772 7420  - SNL - skip on L /= 0
* 7773 3776  - DCA - deposit AC in memory and clear AC. indirect, zero page, addr x176
* 7774 3376  - DCA - deposit AC in memory and clear AC. direct, zero page, addr x176
* 7775 5356  - JMP - jump, direct, zero page, addr x156
*
*************************************************************************************
*
* programmers reference at .. http://www.cs.uiowa.edu/~jones/pdp8/man/index.html
*
*************************************************************************************
